"""
Key bindings which are also known by GNU Readline by the given names.

See: http://www.delorie.com/gnu/docs/readline/rlman_13.html
"""

from __future__ import annotations

from typing import Callable, TypeVar, Union, cast

from prompt_toolkit.document import Document
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding.key_bindings import Binding, key_binding
from prompt_toolkit.key_binding.key_processor import KeyPress, KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.search import SearchDirection
from prompt_toolkit.selection import PasteMode

from .completion import display_completions_like_readline, generate_completions

__all__ = [
    "get_by_name",
]


# Typing.
_Handler = Callable[[KeyPressEvent], None]
_HandlerOrBinding = Union[_Handler, Binding]
_T = TypeVar("_T", bound=_HandlerOrBinding)
E = KeyPressEvent


# Registry that maps the Readline command names to their handlers.
_readline_commands: dict[str, Binding] = {}


def register(name: str) -> Callable[[_T], _T]:
    """
    Store handler in the `_readline_commands` dictionary.
    """

    def decorator(handler: _T) -> _T:
        "`handler` is a callable or Binding."
        if isinstance(handler, Binding):
            _readline_commands[name] = handler
        else:
            _readline_commands[name] = key_binding()(cast(_Handler, handler))

        return handler

    return decorator


def get_by_name(name: str) -> Binding:
    """
    Return the handler for the (Readline) command with the given name.
    """
    try:
        return _readline_commands[name]
    except KeyError as e:
        raise KeyError(f"Unknown Readline command: {name!r}") from e


#
# Commands for moving
# See: http://www.delorie.com/gnu/docs/readline/rlman_14.html
#


@register("beginning-of-buffer")
def beginning_of_buffer(event: E) -> None:
    """
    Move to the start of the buffer.
    """
    buff = event.current_buffer
    buff.cursor_position = 0


@register("end-of-buffer")
def end_of_buffer(event: E) -> None:
    """
    Move to the end of the buffer.
    """
    buff = event.current_buffer
    buff.cursor_position = len(buff.text)


@register("beginning-of-line")
def beginning_of_line(event: E) -> None:
    """
    Move to the start of the current line.
    """
    buff = event.current_buffer
    buff.cursor_position += buff.document.get_start_of_line_position(
        after_whitespace=False
    )


@register("end-of-line")
def end_of_line(event: E) -> None:
    """
    Move to the end of the line.
    """
    buff = event.current_buffer
    buff.cursor_position += buff.document.get_end_of_line_position()


@register("forward-char")
def forward_char(event: E) -> None:
    """
    Move forward a character.
    """
    buff = event.current_buffer
    buff.cursor_position += buff.document.get_cursor_right_position(count=event.arg)


@register("backward-char")
def backward_char(event: E) -> None:
    "Move back a character."
    buff = event.current_buffer
    buff.cursor_position += buff.document.get_cursor_left_position(count=event.arg)


@register("forward-word")
def forward_word(event: E) -> None:
    """
    Move forward to the end of the next word. Words are composed of letters and
    digits.
    """
    buff = event.current_buffer
    pos = buff.document.find_next_word_ending(count=event.arg)

    if pos:
        buff.cursor_position += pos


@register("backward-word")
def backward_word(event: E) -> None:
    """
    Move back to the start of the current or previous word. Words are composed
    of letters and digits.
    """
    buff = event.current_buffer
    pos = buff.document.find_previous_word_beginning(count=event.arg)

    if pos:
        buff.cursor_position += pos


@register("clear-screen")
def clear_screen(event: E) -> None:
    """
    Clear the screen and redraw everything at the top of the screen.
    """
    event.app.renderer.clear()


@register("redraw-current-line")
def redraw_current_line(event: E) -> None:
    """
    Refresh the current line.
    (Readline defines this command, but prompt-toolkit doesn't have it.)
    """
    pass


#
# Commands for manipulating the history.
# See: http://www.delorie.com/gnu/docs/readline/rlman_15.html
#


@register("accept-line")
def accept_line(event: E) -> None:
    """
    Accept the line regardless of where the cursor is.
    """
    event.current_buffer.validate_and_handle()


@register("previous-history")
def previous_history(event: E) -> None:
    """
    Move `back` through the history list, fetching the previous command.
    """
    event.current_buffer.history_backward(count=event.arg)


@register("next-history")
def next_history(event: E) -> None:
    """
    Move `forward` through the history list, fetching the next command.
    """
    event.current_buffer.history_forward(count=event.arg)


@register("beginning-of-history")
def beginning_of_history(event: E) -> None:
    """
    Move to the first line in the history.
    """
    event.current_buffer.go_to_history(0)


@register("end-of-history")
def end_of_history(event: E) -> None:
    """
    Move to the end of the input history, i.e., the line currently being entered.
    """
    event.current_buffer.history_forward(count=10**100)
    buff = event.current_buffer
    buff.go_to_history(len(buff._working_lines) - 1)


@register("reverse-search-history")
def reverse_search_history(event: E) -> None:
    """
    Search backward starting at the current line and moving `up` through
    the history as necessary. This is an incremental search.
    """
    control = event.app.layout.current_control

    if isinstance(control, BufferControl) and control.search_buffer_control:
        event.app.current_search_state.direction = SearchDirection.BACKWARD
        event.app.layout.current_control = control.search_buffer_control


#
# Commands for changing text
#


@register("end-of-file")
def end_of_file(event: E) -> None:
    """
    Exit.
    """
    event.app.exit()


@register("delete-char")
def delete_char(event: E) -> None:
    """
    Delete character before the cursor.
    """
    deleted = event.current_buffer.delete(count=event.arg)
    if not deleted:
        event.app.output.bell()


@register("backward-delete-char")
def backward_delete_char(event: E) -> None:
    """
    Delete the character behind the cursor.
    """
    if event.arg < 0:
        # When a negative argument has been given, this should delete in front
        # of the cursor.
        deleted = event.current_buffer.delete(count=-event.arg)
    else:
        deleted = event.current_buffer.delete_before_cursor(count=event.arg)

    if not deleted:
        event.app.output.bell()


@register("self-insert")
def self_insert(event: E) -> None:
    """
    Insert yourself.
    """
    event.current_buffer.insert_text(event.data * event.arg)


@register("transpose-chars")
def transpose_chars(event: E) -> None:
    """
    Emulate Emacs transpose-char behavior: at the beginning of the buffer,
    do nothing.  At the end of a line or buffer, swap the characters before
    the cursor.  Otherwise, move the cursor right, and then swap the
    characters before the cursor.
    """
    b = event.current_buffer
    p = b.cursor_position
    if p == 0:
        return
    elif p == len(b.text) or b.text[p] == "\n":
        b.swap_characters_before_cursor()
    else:
        b.cursor_position += b.document.get_cursor_right_position()
        b.swap_characters_before_cursor()


@register("uppercase-word")
def uppercase_word(event: E) -> None:
    """
    Uppercase the current (or following) word.
    """
    buff = event.current_buffer

    for i in range(event.arg):
        pos = buff.document.find_next_word_ending()
        words = buff.document.text_after_cursor[:pos]
        buff.insert_text(words.upper(), overwrite=True)


@register("downcase-word")
def downcase_word(event: E) -> None:
    """
    Lowercase the current (or following) word.
    """
    buff = event.current_buffer

    for i in range(event.arg):  # XXX: not DRY: see meta_c and meta_u!!
        pos = buff.document.find_next_word_ending()
        words = buff.document.text_after_cursor[:pos]
        buff.insert_text(words.lower(), overwrite=True)


@register("capitalize-word")
def capitalize_word(event: E) -> None:
    """
    Capitalize the current (or following) word.
    """
    buff = event.current_buffer

    for i in range(event.arg):
        pos = buff.document.find_next_word_ending()
        words = buff.document.text_after_cursor[:pos]
        buff.insert_text(words.title(), overwrite=True)


@register("quoted-insert")
def quoted_insert(event: E) -> None:
    """
    Add the next character typed to the line verbatim. This is how to insert
    key sequences like C-q, for example.
    """
    event.app.quoted_insert = True


#
# Killing and yanking.
#


@register("kill-line")
def kill_line(event: E) -> None:
    """
    Kill the text from the cursor to the end of the line.

    If we are at the end of the line, this should remove the newline.
    (That way, it is possible to delete multiple lines by executing this
    command multiple times.)
    """
    buff = event.current_buffer
    if event.arg < 0:
        deleted = buff.delete_before_cursor(
            count=-buff.document.get_start_of_line_position()
        )
    else:
        if buff.document.current_char == "\n":
            deleted = buff.delete(1)
        else:
            deleted = buff.delete(count=buff.document.get_end_of_line_position())
    event.app.clipboard.set_text(deleted)


@register("kill-word")
def kill_word(event: E) -> None:
    """
    Kill from point to the end of the current word, or if between words, to the
    end of the next word. Word boundaries are the same as forward-word.
    """
    buff = event.current_buffer
    pos = buff.document.find_next_word_ending(count=event.arg)

    if pos:
        deleted = buff.delete(count=pos)

        if event.is_repeat:
            deleted = event.app.clipboard.get_data().text + deleted

        event.app.clipboard.set_text(deleted)


@register("unix-word-rubout")
def unix_word_rubout(event: E, WORD: bool = True) -> None:
    """
    Kill the word behind point, using whitespace as a word boundary.
    Usually bound to ControlW.
    """
    buff = event.current_buffer
    pos = buff.document.find_start_of_previous_word(count=event.arg, WORD=WORD)

    if pos is None:
        # Nothing found? delete until the start of the document.  (The
        # input starts with whitespace and no words were found before the
        # cursor.)
        pos = -buff.cursor_position

    if pos:
        deleted = buff.delete_before_cursor(count=-pos)

        # If the previous key press was also Control-W, concatenate deleted
        # text.
        if event.is_repeat:
            deleted += event.app.clipboard.get_data().text

        event.app.clipboard.set_text(deleted)
    else:
        # Nothing to delete. Bell.
        event.app.output.bell()


@register("backward-kill-word")
def backward_kill_word(event: E) -> None:
    """
    Kills the word before point, using "not a letter nor a digit" as a word boundary.
    Usually bound to M-Del or M-Backspace.
    """
    unix_word_rubout(event, WORD=False)


@register("delete-horizontal-space")
def delete_horizontal_space(event: E) -> None:
    """
    Delete all spaces and tabs around point.
    """
    buff = event.current_buffer
    text_before_cursor = buff.document.text_before_cursor
    text_after_cursor = buff.document.text_after_cursor

    delete_before = len(text_before_cursor) - len(text_before_cursor.rstrip("\t "))
    delete_after = len(text_after_cursor) - len(text_after_cursor.lstrip("\t "))

    buff.delete_before_cursor(count=delete_before)
    buff.delete(count=delete_after)


@register("unix-line-discard")
def unix_line_discard(event: E) -> None:
    """
    Kill backward from the cursor to the beginning of the current line.
    """
    buff = event.current_buffer

    if buff.document.cursor_position_col == 0 and buff.document.cursor_position > 0:
        buff.delete_before_cursor(count=1)
    else:
        deleted = buff.delete_before_cursor(
            count=-buff.document.get_start_of_line_position()
        )
        event.app.clipboard.set_text(deleted)


@register("yank")
def yank(event: E) -> None:
    """
    Paste before cursor.
    """
    event.current_buffer.paste_clipboard_data(
        event.app.clipboard.get_data(), count=event.arg, paste_mode=PasteMode.EMACS
    )


@register("yank-nth-arg")
def yank_nth_arg(event: E) -> None:
    """
    Insert the first argument of the previous command. With an argument, insert
    the nth word from the previous command (start counting at 0).
    """
    n = event.arg if event.arg_present else None
    event.current_buffer.yank_nth_arg(n)


@register("yank-last-arg")
def yank_last_arg(event: E) -> None:
    """
    Like `yank_nth_arg`, but if no argument has been given, yank the last word
    of each line.
    """
    n = event.arg if event.arg_present else None
    event.current_buffer.yank_last_arg(n)


@register("yank-pop")
def yank_pop(event: E) -> None:
    """
    Rotate the kill ring, and yank the new top. Only works following yank or
    yank-pop.
    """
    buff = event.current_buffer
    doc_before_paste = buff.document_before_paste
    clipboard = event.app.clipboard

    if doc_before_paste is not None:
        buff.document = doc_before_paste
        clipboard.rotate()
        buff.paste_clipboard_data(clipboard.get_data(), paste_mode=PasteMode.EMACS)


#
# Completion.
#


@register("complete")
def complete(event: E) -> None:
    """
    Attempt to perform completion.
    """
    display_completions_like_readline(event)


@register("menu-complete")
def menu_complete(event: E) -> None:
    """
    Generate completions, or go to the next completion. (This is the default
    way of completing input in prompt_toolkit.)
    """
    generate_completions(event)


@register("menu-complete-backward")
def menu_complete_backward(event: E) -> None:
    """
    Move backward through the list of possible completions.
    """
    event.current_buffer.complete_previous()


#
# Keyboard macros.
#


@register("start-kbd-macro")
def start_kbd_macro(event: E) -> None:
    """
    Begin saving the characters typed into the current keyboard macro.
    """
    event.app.emacs_state.start_macro()


@register("end-kbd-macro")
def end_kbd_macro(event: E) -> None:
    """
    Stop saving the characters typed into the current keyboard macro and save
    the definition.
    """
    event.app.emacs_state.end_macro()


@register("call-last-kbd-macro")
@key_binding(record_in_macro=False)
def call_last_kbd_macro(event: E) -> None:
    """
    Re-execute the last keyboard macro defined, by making the characters in the
    macro appear as if typed at the keyboard.

    Notice that we pass `record_in_macro=False`. This ensures that the 'c-x e'
    key sequence doesn't appear in the recording itself. This function inserts
    the body of the called macro back into the KeyProcessor, so these keys will
    be added later on to the macro of their handlers have `record_in_macro=True`.
    """
    # Insert the macro.
    macro = event.app.emacs_state.macro

    if macro:
        event.app.key_processor.feed_multiple(macro, first=True)


@register("print-last-kbd-macro")
def print_last_kbd_macro(event: E) -> None:
    """
    Print the last keyboard macro.
    """

    # TODO: Make the format suitable for the inputrc file.
    def print_macro() -> None:
        macro = event.app.emacs_state.macro
        if macro:
            for k in macro:
                print(k)

    from prompt_toolkit.application.run_in_terminal import run_in_terminal

    run_in_terminal(print_macro)


#
# Miscellaneous Commands.
#


@register("undo")
def undo(event: E) -> None:
    """
    Incremental undo.
    """
    event.current_buffer.undo()


@register("insert-comment")
def insert_comment(event: E) -> None:
    """
    Without numeric argument, comment all lines.
    With numeric argument, uncomment all lines.
    In any case accept the input.
    """
    buff = event.current_buffer

    # Transform all lines.
    if event.arg != 1:

        def change(line: str) -> str:
            return line[1:] if line.startswith("#") else line

    else:

        def change(line: str) -> str:
            return "#" + line

    buff.document = Document(
        text="\n".join(map(change, buff.text.splitlines())), cursor_position=0
    )

    # Accept input.
    buff.validate_and_handle()


@register("vi-editing-mode")
def vi_editing_mode(event: E) -> None:
    """
    Switch to Vi editing mode.
    """
    event.app.editing_mode = EditingMode.VI


@register("emacs-editing-mode")
def emacs_editing_mode(event: E) -> None:
    """
    Switch to Emacs editing mode.
    """
    event.app.editing_mode = EditingMode.EMACS


@register("prefix-meta")
def prefix_meta(event: E) -> None:
    """
    Metafy the next character typed. This is for keyboards without a meta key.

    Sometimes people also want to bind other keys to Meta, e.g. 'jj'::

        key_bindings.add_key_binding('j', 'j', filter=ViInsertMode())(prefix_meta)
    """
    # ('first' should be true, because we want to insert it at the current
    # position in the queue.)
    event.app.key_processor.feed(KeyPress(Keys.Escape), first=True)


@register("operate-and-get-next")
def operate_and_get_next(event: E) -> None:
    """
    Accept the current line for execution and fetch the next line relative to
    the current line from the history for editing.
    """
    buff = event.current_buffer
    new_index = buff.working_index + 1

    # Accept the current input. (This will also redraw the interface in the
    # 'done' state.)
    buff.validate_and_handle()

    # Set the new index at the start of the next run.
    def set_working_index() -> None:
        if new_index < len(buff._working_lines):
            buff.working_index = new_index

    event.app.pre_run_callables.append(set_working_index)


@register("edit-and-execute-command")
def edit_and_execute(event: E) -> None:
    """
    Invoke an editor on the current command line, and accept the result.
    """
    buff = event.current_buffer
    buff.open_in_editor(validate_and_handle=True)



"""
Base classes for prompt_toolkit lexers.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable, Hashable

from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text.base import StyleAndTextTuples

__all__ = [
    "Lexer",
    "SimpleLexer",
    "DynamicLexer",
]


class Lexer(metaclass=ABCMeta):
    """
    Base class for all lexers.
    """

    @abstractmethod
    def lex_document(self, document: Document) -> Callable[[int], StyleAndTextTuples]:
        """
        Takes a :class:`~prompt_toolkit.document.Document` and returns a
        callable that takes a line number and returns a list of
        ``(style_str, text)`` tuples for that line.

        XXX: Note that in the past, this was supposed to return a list
             of ``(Token, text)`` tuples, just like a Pygments lexer.
        """

    def invalidation_hash(self) -> Hashable:
        """
        When this changes, `lex_document` could give a different output.
        (Only used for `DynamicLexer`.)
        """
        return id(self)


class SimpleLexer(Lexer):
    """
    Lexer that doesn't do any tokenizing and returns the whole input as one
    token.

    :param style: The style string for this lexer.
    """

    def __init__(self, style: str = "") -> None:
        self.style = style

    def lex_document(self, document: Document) -> Callable[[int], StyleAndTextTuples]:
        lines = document.lines

        def get_line(lineno: int) -> StyleAndTextTuples:
            "Return the tokens for the given line."
            try:
                return [(self.style, lines[lineno])]
            except IndexError:
                return []

        return get_line


class DynamicLexer(Lexer):
    """
    Lexer class that can dynamically returns any Lexer.

    :param get_lexer: Callable that returns a :class:`.Lexer` instance.
    """

    def __init__(self, get_lexer: Callable[[], Lexer | None]) -> None:
        self.get_lexer = get_lexer
        self._dummy = SimpleLexer()

    def lex_document(self, document: Document) -> Callable[[int], StyleAndTextTuples]:
        lexer = self.get_lexer() or self._dummy
        return lexer.lex_document(document)

    def invalidation_hash(self) -> Hashable:
        lexer = self.get_lexer() or self._dummy
        return id(lexer)



"""Abstract base class for MIDI backends."""

from abc import ABC, abstractmethod
from pathlib import Path

from ..types import MidiSequence


class MidiBackend(ABC):
    """Abstract base class for MIDI output backends."""

    @abstractmethod
    def play(self, sequence: MidiSequence) -> None:
        """Play a MIDI sequence in realtime.

        Args:
            sequence: The MIDI sequence to play.
        """
        pass

    @abstractmethod
    def save(self, sequence: MidiSequence, path: Path | str) -> None:
        """Save a MIDI sequence to a file.

        Args:
            sequence: The MIDI sequence to save.
            path: The output file path.
        """
        pass

    @abstractmethod
    def stop(self) -> None:
        """Stop any currently playing sequence."""
        pass

    def is_playing(self) -> bool:
        """Check if a sequence is currently playing.

        Returns:
            True if playing, False otherwise.
        """
        return False



"""
Output for vt100 terminals.

A lot of thanks, regarding outputting of colors, goes to the Pygments project:
(We don't rely on Pygments anymore, because many things are very custom, and
everything has been highly optimized.)
http://pygments.org/
"""

from __future__ import annotations

import io
import os
import sys
from typing import Callable, Dict, Hashable, Iterable, Sequence, TextIO, Tuple

from prompt_toolkit.cursor_shapes import CursorShape
from prompt_toolkit.data_structures import Size
from prompt_toolkit.output import Output
from prompt_toolkit.styles import ANSI_COLOR_NAMES, Attrs
from prompt_toolkit.utils import is_dumb_terminal

from .color_depth import ColorDepth
from .flush_stdout import flush_stdout

__all__ = [
    "Vt100_Output",
]


FG_ANSI_COLORS = {
    "ansidefault": 39,
    # Low intensity.
    "ansiblack": 30,
    "ansired": 31,
    "ansigreen": 32,
    "ansiyellow": 33,
    "ansiblue": 34,
    "ansimagenta": 35,
    "ansicyan": 36,
    "ansigray": 37,
    # High intensity.
    "ansibrightblack": 90,
    "ansibrightred": 91,
    "ansibrightgreen": 92,
    "ansibrightyellow": 93,
    "ansibrightblue": 94,
    "ansibrightmagenta": 95,
    "ansibrightcyan": 96,
    "ansiwhite": 97,
}

BG_ANSI_COLORS = {
    "ansidefault": 49,
    # Low intensity.
    "ansiblack": 40,
    "ansired": 41,
    "ansigreen": 42,
    "ansiyellow": 43,
    "ansiblue": 44,
    "ansimagenta": 45,
    "ansicyan": 46,
    "ansigray": 47,
    # High intensity.
    "ansibrightblack": 100,
    "ansibrightred": 101,
    "ansibrightgreen": 102,
    "ansibrightyellow": 103,
    "ansibrightblue": 104,
    "ansibrightmagenta": 105,
    "ansibrightcyan": 106,
    "ansiwhite": 107,
}


ANSI_COLORS_TO_RGB = {
    "ansidefault": (
        0x00,
        0x00,
        0x00,
    ),  # Don't use, 'default' doesn't really have a value.
    "ansiblack": (0x00, 0x00, 0x00),
    "ansigray": (0xE5, 0xE5, 0xE5),
    "ansibrightblack": (0x7F, 0x7F, 0x7F),
    "ansiwhite": (0xFF, 0xFF, 0xFF),
    # Low intensity.
    "ansired": (0xCD, 0x00, 0x00),
    "ansigreen": (0x00, 0xCD, 0x00),
    "ansiyellow": (0xCD, 0xCD, 0x00),
    "ansiblue": (0x00, 0x00, 0xCD),
    "ansimagenta": (0xCD, 0x00, 0xCD),
    "ansicyan": (0x00, 0xCD, 0xCD),
    # High intensity.
    "ansibrightred": (0xFF, 0x00, 0x00),
    "ansibrightgreen": (0x00, 0xFF, 0x00),
    "ansibrightyellow": (0xFF, 0xFF, 0x00),
    "ansibrightblue": (0x00, 0x00, 0xFF),
    "ansibrightmagenta": (0xFF, 0x00, 0xFF),
    "ansibrightcyan": (0x00, 0xFF, 0xFF),
}


assert set(FG_ANSI_COLORS) == set(ANSI_COLOR_NAMES)
assert set(BG_ANSI_COLORS) == set(ANSI_COLOR_NAMES)
assert set(ANSI_COLORS_TO_RGB) == set(ANSI_COLOR_NAMES)


def _get_closest_ansi_color(r: int, g: int, b: int, exclude: Sequence[str] = ()) -> str:
    """
    Find closest ANSI color. Return it by name.

    :param r: Red (Between 0 and 255.)
    :param g: Green (Between 0 and 255.)
    :param b: Blue (Between 0 and 255.)
    :param exclude: A tuple of color names to exclude. (E.g. ``('ansired', )``.)
    """
    exclude = list(exclude)

    # When we have a bit of saturation, avoid the gray-like colors, otherwise,
    # too often the distance to the gray color is less.
    saturation = abs(r - g) + abs(g - b) + abs(b - r)  # Between 0..510

    if saturation > 30:
        exclude.extend(["ansilightgray", "ansidarkgray", "ansiwhite", "ansiblack"])

    # Take the closest color.
    # (Thanks to Pygments for this part.)
    distance = 257 * 257 * 3  # "infinity" (>distance from #000000 to #ffffff)
    match = "ansidefault"

    for name, (r2, g2, b2) in ANSI_COLORS_TO_RGB.items():
        if name != "ansidefault" and name not in exclude:
            d = (r - r2) ** 2 + (g - g2) ** 2 + (b - b2) ** 2

            if d < distance:
                match = name
                distance = d

    return match


_ColorCodeAndName = Tuple[int, str]


class _16ColorCache:
    """
    Cache which maps (r, g, b) tuples to 16 ansi colors.

    :param bg: Cache for background colors, instead of foreground.
    """

    def __init__(self, bg: bool = False) -> None:
        self.bg = bg
        self._cache: dict[Hashable, _ColorCodeAndName] = {}

    def get_code(
        self, value: tuple[int, int, int], exclude: Sequence[str] = ()
    ) -> _ColorCodeAndName:
        """
        Return a (ansi_code, ansi_name) tuple. (E.g. ``(44, 'ansiblue')``.) for
        a given (r,g,b) value.
        """
        key: Hashable = (value, tuple(exclude))
        cache = self._cache

        if key not in cache:
            cache[key] = self._get(value, exclude)

        return cache[key]

    def _get(
        self, value: tuple[int, int, int], exclude: Sequence[str] = ()
    ) -> _ColorCodeAndName:
        r, g, b = value
        match = _get_closest_ansi_color(r, g, b, exclude=exclude)

        # Turn color name into code.
        if self.bg:
            code = BG_ANSI_COLORS[match]
        else:
            code = FG_ANSI_COLORS[match]

        return code, match


class _256ColorCache(Dict[Tuple[int, int, int], int]):
    """
    Cache which maps (r, g, b) tuples to 256 colors.
    """

    def __init__(self) -> None:
        # Build color table.
        colors: list[tuple[int, int, int]] = []

        # colors 0..15: 16 basic colors
        colors.append((0x00, 0x00, 0x00))  # 0
        colors.append((0xCD, 0x00, 0x00))  # 1
        colors.append((0x00, 0xCD, 0x00))  # 2
        colors.append((0xCD, 0xCD, 0x00))  # 3
        colors.append((0x00, 0x00, 0xEE))  # 4
        colors.append((0xCD, 0x00, 0xCD))  # 5
        colors.append((0x00, 0xCD, 0xCD))  # 6
        colors.append((0xE5, 0xE5, 0xE5))  # 7
        colors.append((0x7F, 0x7F, 0x7F))  # 8
        colors.append((0xFF, 0x00, 0x00))  # 9
        colors.append((0x00, 0xFF, 0x00))  # 10
        colors.append((0xFF, 0xFF, 0x00))  # 11
        colors.append((0x5C, 0x5C, 0xFF))  # 12
        colors.append((0xFF, 0x00, 0xFF))  # 13
        colors.append((0x00, 0xFF, 0xFF))  # 14
        colors.append((0xFF, 0xFF, 0xFF))  # 15

        # colors 16..232: the 6x6x6 color cube
        valuerange = (0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF)

        for i in range(217):
            r = valuerange[(i // 36) % 6]
            g = valuerange[(i // 6) % 6]
            b = valuerange[i % 6]
            colors.append((r, g, b))

        # colors 233..253: grayscale
        for i in range(1, 22):
            v = 8 + i * 10
            colors.append((v, v, v))

        self.colors = colors

    def __missing__(self, value: tuple[int, int, int]) -> int:
        r, g, b = value

        # Find closest color.
        # (Thanks to Pygments for this!)
        distance = 257 * 257 * 3  # "infinity" (>distance from #000000 to #ffffff)
        match = 0

        for i, (r2, g2, b2) in enumerate(self.colors):
            if i >= 16:  # XXX: We ignore the 16 ANSI colors when mapping RGB
                # to the 256 colors, because these highly depend on
                # the color scheme of the terminal.
                d = (r - r2) ** 2 + (g - g2) ** 2 + (b - b2) ** 2

                if d < distance:
                    match = i
                    distance = d

        # Turn color name into code.
        self[value] = match
        return match


_16_fg_colors = _16ColorCache(bg=False)
_16_bg_colors = _16ColorCache(bg=True)
_256_colors = _256ColorCache()


class _EscapeCodeCache(Dict[Attrs, str]):
    """
    Cache for VT100 escape codes. It maps
    (fgcolor, bgcolor, bold, underline, strike, italic, blink, reverse, hidden, dim) tuples to VT100
    escape sequences.

    :param true_color: When True, use 24bit colors instead of 256 colors.
    """

    def __init__(self, color_depth: ColorDepth) -> None:
        self.color_depth = color_depth

    def __missing__(self, attrs: Attrs) -> str:
        (
            fgcolor,
            bgcolor,
            bold,
            underline,
            strike,
            italic,
            blink,
            reverse,
            hidden,
            dim,
        ) = attrs
        parts: list[str] = []

        parts.extend(self._colors_to_code(fgcolor or "", bgcolor or ""))

        if bold:
            parts.append("1")
        if dim:
            parts.append("2")
        if italic:
            parts.append("3")
        if blink:
            parts.append("5")
        if underline:
            parts.append("4")
        if reverse:
            parts.append("7")
        if hidden:
            parts.append("8")
        if strike:
            parts.append("9")

        if parts:
            result = "\x1b[0;" + ";".join(parts) + "m"
        else:
            result = "\x1b[0m"

        self[attrs] = result
        return result

    def _color_name_to_rgb(self, color: str) -> tuple[int, int, int]:
        "Turn 'ffffff', into (0xff, 0xff, 0xff)."
        try:
            rgb = int(color, 16)
        except ValueError:
            raise
        else:
            r = (rgb >> 16) & 0xFF
            g = (rgb >> 8) & 0xFF
            b = rgb & 0xFF
            return r, g, b

    def _colors_to_code(self, fg_color: str, bg_color: str) -> Iterable[str]:
        """
        Return a tuple with the vt100 values  that represent this color.
        """
        # When requesting ANSI colors only, and both fg/bg color were converted
        # to ANSI, ensure that the foreground and background color are not the
        # same. (Unless they were explicitly defined to be the same color.)
        fg_ansi = ""

        def get(color: str, bg: bool) -> list[int]:
            nonlocal fg_ansi

            table = BG_ANSI_COLORS if bg else FG_ANSI_COLORS

            if not color or self.color_depth == ColorDepth.DEPTH_1_BIT:
                return []

            # 16 ANSI colors. (Given by name.)
            elif color in table:
                return [table[color]]

            # RGB colors. (Defined as 'ffffff'.)
            else:
                try:
                    rgb = self._color_name_to_rgb(color)
                except ValueError:
                    return []

                # When only 16 colors are supported, use that.
                if self.color_depth == ColorDepth.DEPTH_4_BIT:
                    if bg:  # Background.
                        if fg_color != bg_color:
                            exclude = [fg_ansi]
                        else:
                            exclude = []
                        code, name = _16_bg_colors.get_code(rgb, exclude=exclude)
                        return [code]
                    else:  # Foreground.
                        code, name = _16_fg_colors.get_code(rgb)
                        fg_ansi = name
                        return [code]

                # True colors. (Only when this feature is enabled.)
                elif self.color_depth == ColorDepth.DEPTH_24_BIT:
                    r, g, b = rgb
                    return [(48 if bg else 38), 2, r, g, b]

                # 256 RGB colors.
                else:
                    return [(48 if bg else 38), 5, _256_colors[rgb]]

        result: list[int] = []
        result.extend(get(fg_color, False))
        result.extend(get(bg_color, True))

        return map(str, result)


def _get_size(fileno: int) -> tuple[int, int]:
    """
    Get the size of this pseudo terminal.

    :param fileno: stdout.fileno()
    :returns: A (rows, cols) tuple.
    """
    size = os.get_terminal_size(fileno)
    return size.lines, size.columns


class Vt100_Output(Output):
    """
    :param get_size: A callable which returns the `Size` of the output terminal.
    :param stdout: Any object with has a `write` and `flush` method + an 'encoding' property.
    :param term: The terminal environment variable. (xterm, xterm-256color, linux, ...)
    :param enable_cpr: When `True` (the default), send "cursor position
        request" escape sequences to the output in order to detect the cursor
        position. That way, we can properly determine how much space there is
        available for the UI (especially for drop down menus) to render. The
        `Renderer` will still try to figure out whether the current terminal
        does respond to CPR escapes. When `False`, never attempt to send CPR
        requests.
    """

    # For the error messages. Only display "Output is not a terminal" once per
    # file descriptor.
    _fds_not_a_terminal: set[int] = set()

    def __init__(
        self,
        stdout: TextIO,
        get_size: Callable[[], Size],
        term: str | None = None,
        default_color_depth: ColorDepth | None = None,
        enable_bell: bool = True,
        enable_cpr: bool = True,
    ) -> None:
        assert all(hasattr(stdout, a) for a in ("write", "flush"))

        self._buffer: list[str] = []
        self.stdout: TextIO = stdout
        self.default_color_depth = default_color_depth
        self._get_size = get_size
        self.term = term
        self.enable_bell = enable_bell
        self.enable_cpr = enable_cpr

        # Cache for escape codes.
        self._escape_code_caches: dict[ColorDepth, _EscapeCodeCache] = {
            ColorDepth.DEPTH_1_BIT: _EscapeCodeCache(ColorDepth.DEPTH_1_BIT),
            ColorDepth.DEPTH_4_BIT: _EscapeCodeCache(ColorDepth.DEPTH_4_BIT),
            ColorDepth.DEPTH_8_BIT: _EscapeCodeCache(ColorDepth.DEPTH_8_BIT),
            ColorDepth.DEPTH_24_BIT: _EscapeCodeCache(ColorDepth.DEPTH_24_BIT),
        }

        # Keep track of whether the cursor shape was ever changed.
        # (We don't restore the cursor shape if it was never changed - by
        # default, we don't change them.)
        self._cursor_shape_changed = False

        # Don't hide/show the cursor when this was already done.
        # (`None` means that we don't know whether the cursor is visible or
        # not.)
        self._cursor_visible: bool | None = None

    @classmethod
    def from_pty(
        cls,
        stdout: TextIO,
        term: str | None = None,
        default_color_depth: ColorDepth | None = None,
        enable_bell: bool = True,
    ) -> Vt100_Output:
        """
        Create an Output class from a pseudo terminal.
        (This will take the dimensions by reading the pseudo
        terminal attributes.)
        """
        fd: int | None
        # Normally, this requires a real TTY device, but people instantiate
        # this class often during unit tests as well. For convenience, we print
        # an error message, use standard dimensions, and go on.
        try:
            fd = stdout.fileno()
        except io.UnsupportedOperation:
            fd = None

        if not stdout.isatty() and (fd is None or fd not in cls._fds_not_a_terminal):
            msg = "Warning: Output is not a terminal (fd=%r).\n"
            sys.stderr.write(msg % fd)
            sys.stderr.flush()
            if fd is not None:
                cls._fds_not_a_terminal.add(fd)

        def get_size() -> Size:
            # If terminal (incorrectly) reports its size as 0, pick a
            # reasonable default.  See
            # https://github.com/ipython/ipython/issues/10071
            rows, columns = (None, None)

            # It is possible that `stdout` is no longer a TTY device at this
            # point. In that case we get an `OSError` in the ioctl call in
            # `get_size`. See:
            # https://github.com/prompt-toolkit/python-prompt-toolkit/pull/1021
            try:
                rows, columns = _get_size(stdout.fileno())
            except OSError:
                pass
            return Size(rows=rows or 24, columns=columns or 80)

        return cls(
            stdout,
            get_size,
            term=term,
            default_color_depth=default_color_depth,
            enable_bell=enable_bell,
        )

    def get_size(self) -> Size:
        return self._get_size()

    def fileno(self) -> int:
        "Return file descriptor."
        return self.stdout.fileno()

    def encoding(self) -> str:
        "Return encoding used for stdout."
        return self.stdout.encoding

    def write_raw(self, data: str) -> None:
        """
        Write raw data to output.
        """
        self._buffer.append(data)

    def write(self, data: str) -> None:
        """
        Write text to output.
        (Removes vt100 escape codes. -- used for safely writing text.)
        """
        self._buffer.append(data.replace("\x1b", "?"))

    def set_title(self, title: str) -> None:
        """
        Set terminal title.
        """
        if self.term not in (
            "linux",
            "eterm-color",
        ):  # Not supported by the Linux console.
            self.write_raw(
                "\x1b]2;{}\x07".format(title.replace("\x1b", "").replace("\x07", ""))
            )

    def clear_title(self) -> None:
        self.set_title("")

    def erase_screen(self) -> None:
        """
        Erases the screen with the background color and moves the cursor to
        home.
        """
        self.write_raw("\x1b[2J")

    def enter_alternate_screen(self) -> None:
        self.write_raw("\x1b[?1049h\x1b[H")

    def quit_alternate_screen(self) -> None:
        self.write_raw("\x1b[?1049l")

    def enable_mouse_support(self) -> None:
        self.write_raw("\x1b[?1000h")

        # Enable mouse-drag support.
        self.write_raw("\x1b[?1003h")

        # Enable urxvt Mouse mode. (For terminals that understand this.)
        self.write_raw("\x1b[?1015h")

        # Also enable Xterm SGR mouse mode. (For terminals that understand this.)
        self.write_raw("\x1b[?1006h")

        # Note: E.g. lxterminal understands 1000h, but not the urxvt or sgr
        #       extensions.

    def disable_mouse_support(self) -> None:
        self.write_raw("\x1b[?1000l")
        self.write_raw("\x1b[?1015l")
        self.write_raw("\x1b[?1006l")
        self.write_raw("\x1b[?1003l")

    def erase_end_of_line(self) -> None:
        """
        Erases from the current cursor position to the end of the current line.
        """
        self.write_raw("\x1b[K")

    def erase_down(self) -> None:
        """
        Erases the screen from the current line down to the bottom of the
        screen.
        """
        self.write_raw("\x1b[J")

    def reset_attributes(self) -> None:
        self.write_raw("\x1b[0m")

    def set_attributes(self, attrs: Attrs, color_depth: ColorDepth) -> None:
        """
        Create new style and output.

        :param attrs: `Attrs` instance.
        """
        # Get current depth.
        escape_code_cache = self._escape_code_caches[color_depth]

        # Write escape character.
        self.write_raw(escape_code_cache[attrs])

    def disable_autowrap(self) -> None:
        self.write_raw("\x1b[?7l")

    def enable_autowrap(self) -> None:
        self.write_raw("\x1b[?7h")

    def enable_bracketed_paste(self) -> None:
        self.write_raw("\x1b[?2004h")

    def disable_bracketed_paste(self) -> None:
        self.write_raw("\x1b[?2004l")

    def reset_cursor_key_mode(self) -> None:
        """
        For vt100 only.
        Put the terminal in cursor mode (instead of application mode).
        """
        # Put the terminal in cursor mode. (Instead of application mode.)
        self.write_raw("\x1b[?1l")

    def cursor_goto(self, row: int = 0, column: int = 0) -> None:
        """
        Move cursor position.
        """
        self.write_raw("\x1b[%i;%iH" % (row, column))

    def cursor_up(self, amount: int) -> None:
        if amount == 0:
            pass
        elif amount == 1:
            self.write_raw("\x1b[A")
        else:
            self.write_raw("\x1b[%iA" % amount)

    def cursor_down(self, amount: int) -> None:
        if amount == 0:
            pass
        elif amount == 1:
            # Note: Not the same as '\n', '\n' can cause the window content to
            #       scroll.
            self.write_raw("\x1b[B")
        else:
            self.write_raw("\x1b[%iB" % amount)

    def cursor_forward(self, amount: int) -> None:
        if amount == 0:
            pass
        elif amount == 1:
            self.write_raw("\x1b[C")
        else:
            self.write_raw("\x1b[%iC" % amount)

    def cursor_backward(self, amount: int) -> None:
        if amount == 0:
            pass
        elif amount == 1:
            self.write_raw("\b")  # '\x1b[D'
        else:
            self.write_raw("\x1b[%iD" % amount)

    def hide_cursor(self) -> None:
        if self._cursor_visible in (True, None):
            self._cursor_visible = False
            self.write_raw("\x1b[?25l")

    def show_cursor(self) -> None:
        if self._cursor_visible in (False, None):
            self._cursor_visible = True
            self.write_raw("\x1b[?12l\x1b[?25h")  # Stop blinking cursor and show.

    def set_cursor_shape(self, cursor_shape: CursorShape) -> None:
        if cursor_shape == CursorShape._NEVER_CHANGE:
            return

        self._cursor_shape_changed = True
        self.write_raw(
            {
                CursorShape.BLOCK: "\x1b[2 q",
                CursorShape.BEAM: "\x1b[6 q",
                CursorShape.UNDERLINE: "\x1b[4 q",
                CursorShape.BLINKING_BLOCK: "\x1b[1 q",
                CursorShape.BLINKING_BEAM: "\x1b[5 q",
                CursorShape.BLINKING_UNDERLINE: "\x1b[3 q",
            }.get(cursor_shape, "")
        )

    def reset_cursor_shape(self) -> None:
        "Reset cursor shape."
        # (Only reset cursor shape, if we ever changed it.)
        if self._cursor_shape_changed:
            self._cursor_shape_changed = False

            # Reset cursor shape.
            self.write_raw("\x1b[0 q")

    def flush(self) -> None:
        """
        Write to output stream and flush.
        """
        if not self._buffer:
            return

        data = "".join(self._buffer)
        self._buffer = []

        flush_stdout(self.stdout, data)

    def ask_for_cpr(self) -> None:
        """
        Asks for a cursor position report (CPR).
        """
        self.write_raw("\x1b[6n")
        self.flush()

    @property
    def responds_to_cpr(self) -> bool:
        if not self.enable_cpr:
            return False

        # When the input is a tty, we assume that CPR is supported.
        # It's not when the input is piped from Pexpect.
        if os.environ.get("PROMPT_TOOLKIT_NO_CPR", "") == "1":
            return False

        if is_dumb_terminal(self.term):
            return False
        try:
            return self.stdout.isatty()
        except ValueError:
            return False  # ValueError: I/O operation on closed file

    def bell(self) -> None:
        "Sound bell."
        if self.enable_bell:
            self.write_raw("\a")
            self.flush()

    def get_default_color_depth(self) -> ColorDepth:
        """
        Return the default color depth for a vt100 terminal, according to the
        our term value.

        We prefer 256 colors almost always, because this is what most terminals
        support these days, and is a good default.
        """
        if self.default_color_depth is not None:
            return self.default_color_depth

        term = self.term

        if term is None:
            return ColorDepth.DEFAULT

        if is_dumb_terminal(term):
            return ColorDepth.DEPTH_1_BIT

        if term in ("linux", "eterm-color"):
            return ColorDepth.DEPTH_4_BIT

        return ColorDepth.DEFAULT



"""Real-time MIDI input transcription."""

from __future__ import annotations

import time
from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Callable, Literal

from .._libremidi import (  # type: ignore[import-not-found]
    MidiIn,
    MidiMessage,
    Observer,
)
from ..compose.attributes import Tempo
from ..compose.core import Note, Rest, Seq, Cram, Chord
from ..compose.part import Part
from ..score import Score
from .midi_to_ast import (
    beats_to_duration,
    duration_value_to_beats,
    midi_pitch_to_note,
)

if TYPE_CHECKING:
    from ..score import Score

# -----------------------------------------------------------------------------
# Transcription timing constants
# -----------------------------------------------------------------------------

# Maximum time difference (seconds) for notes to be grouped into a chord.
# Notes starting within this window are considered simultaneous.
CHORD_GROUPING_TOLERANCE_SECONDS = 0.002

# Minimum gap (seconds) between notes before inserting a rest.
# Gaps smaller than this are absorbed into the preceding note.
MIN_REST_GAP_SECONDS = 0.05

# Swing detection range (beats). Notes with durations in this range
# are candidates for swing quantization (long/short alternation).
SWING_DETECTION_MIN_BEATS = 0.15
SWING_DETECTION_MAX_BEATS = 0.85


@dataclass
class PendingNote:
    """A note that has been started but not yet released."""

    pitch: int
    velocity: int
    start_time: float  # In seconds


@dataclass
class RecordedNote:
    """A completed note with start time and duration."""

    pitch: int
    velocity: int
    start_time: float
    duration: float


@dataclass
class TranscribeSession:
    """A real-time MIDI transcription session.

    Records incoming MIDI notes and converts them to aldakit compose elements.

    Example:
        >>> session = TranscribeSession()
        >>> session.start()  # Opens MIDI input
        >>> # Play some notes on your MIDI keyboard...
        >>> time.sleep(5)
        >>> score = session.stop()  # Returns a Score with recorded notes
        >>> print(score.to_alda())
    """

    port_name: str | None = None
    quantize_grid: float = 0.25  # Quantize to 16th notes
    default_tempo: float = 120.0
    feel: Literal["straight", "swing", "triplet", "quintuplet"] = "straight"
    swing_ratio: float = 2.0 / 3.0  # Portion of beat for the long swing note

    # Internal state
    _midi_in: MidiIn | None = field(default=None, repr=False)
    _pending_notes: dict[int, PendingNote] = field(default_factory=dict, repr=False)
    _recorded_notes: list[RecordedNote] = field(default_factory=list, repr=False)
    _start_time: float = field(default=0.0, repr=False)
    _running: bool = field(default=False, repr=False)
    _on_note: Callable[[int, int, bool], None] | None = field(default=None, repr=False)
    _swing_next_is_long: bool = field(default=True, repr=False)

    def list_input_ports(self) -> list[str]:
        """List available MIDI input ports."""
        observer = Observer()
        return [p.display_name for p in observer.get_input_ports()]

    def start(self, port_name: str | None = None) -> None:
        """Start recording MIDI input.

        Args:
            port_name: MIDI input port name. If None, uses the first available port.

        Raises:
            RuntimeError: If no MIDI input ports are available.
        """
        if self._running:
            return

        self._midi_in = MidiIn()
        observer = Observer()
        input_ports = observer.get_input_ports()

        if not input_ports:
            raise RuntimeError("No MIDI input ports available")

        # Find the requested port or use the first one
        target_port = None
        if port_name:
            for port in input_ports:
                if port_name.lower() in port.display_name.lower():
                    target_port = port
                    break
            if target_port is None:
                raise RuntimeError(f"MIDI input port '{port_name}' not found")
        else:
            target_port = input_ports[0]

        err = self._midi_in.open_port(target_port)
        if err:
            raise RuntimeError(f"Failed to open MIDI port: {err}")

        self._pending_notes = {}
        self._recorded_notes = []
        self._start_time = time.time()
        self._running = True
        self._swing_next_is_long = True

    def stop(self) -> Seq:
        """Stop recording and return the recorded notes as a Seq.

        Returns:
            A Seq containing the recorded notes.
        """
        if not self._running:
            return Seq()

        # Process any remaining messages
        self.poll()

        # Close any pending notes
        end_time = time.time() - self._start_time
        for pitch, pending in self._pending_notes.items():
            duration = end_time - pending.start_time
            self._recorded_notes.append(
                RecordedNote(
                    pitch=pending.pitch,
                    velocity=pending.velocity,
                    start_time=pending.start_time,
                    duration=max(0.1, duration),
                )
            )

        self._running = False
        if self._midi_in:
            self._midi_in.close_port()
            self._midi_in = None
        self._swing_next_is_long = True

        # Convert recorded notes to Seq
        return self._notes_to_seq()

    def poll(self) -> None:
        """Poll for incoming MIDI messages. Call this periodically."""
        if not self._running or not self._midi_in:
            return

        current_time = time.time() - self._start_time
        messages = self._midi_in.poll()

        for msg in messages:
            self._process_message(msg, current_time)

    def _process_message(self, msg: MidiMessage, current_time: float) -> None:
        """Process a single MIDI message."""
        if len(msg.bytes) < 2:
            return

        status = msg.bytes[0]
        msg_type = status & 0xF0

        if msg_type == 0x90 and len(msg.bytes) >= 3:
            # Note On
            pitch = msg.bytes[1]
            velocity = msg.bytes[2]

            if velocity == 0:
                # Note On with velocity 0 = Note Off
                self._note_off(pitch, current_time)
            else:
                self._note_on(pitch, velocity, current_time)

        elif msg_type == 0x80 and len(msg.bytes) >= 3:
            # Note Off
            pitch = msg.bytes[1]
            self._note_off(pitch, current_time)

    def _note_on(self, pitch: int, velocity: int, time: float) -> None:
        """Handle a note on event."""
        # If there's already a pending note at this pitch, end it first
        if pitch in self._pending_notes:
            self._note_off(pitch, time)

        self._pending_notes[pitch] = PendingNote(
            pitch=pitch, velocity=velocity, start_time=time
        )

        if self._on_note:
            self._on_note(pitch, velocity, True)

    def _note_off(self, pitch: int, time: float) -> None:
        """Handle a note off event."""
        if pitch not in self._pending_notes:
            return

        pending = self._pending_notes.pop(pitch)
        duration = time - pending.start_time

        self._recorded_notes.append(
            RecordedNote(
                pitch=pending.pitch,
                velocity=pending.velocity,
                start_time=pending.start_time,
                duration=max(0.01, duration),  # Minimum duration
            )
        )

        if self._on_note:
            self._on_note(pitch, 0, False)

    def _notes_to_seq(self) -> Seq:
        """Convert recorded notes to a Seq."""
        if not self._recorded_notes:
            return Seq()

        sorted_notes = sorted(self._recorded_notes, key=lambda n: n.start_time)
        groups = self._group_notes(sorted_notes)
        elements: list = []
        current_time = 0.0

        for group in groups:
            start_time = group[0].start_time
            gap_seconds = start_time - current_time
            if gap_seconds > MIN_REST_GAP_SECONDS:
                gap_beats = self._seconds_to_beats(gap_seconds)
                rest_segments = self._segments_for_beats(gap_beats, kind="rest")
                gained = self._append_rest_segments(rest_segments, elements)
                current_time += self._beats_to_seconds(gained)

            duration_seconds = max(n.duration for n in group)
            duration_beats = self._seconds_to_beats(duration_seconds)
            note_segments = self._segments_for_beats(duration_beats, kind="note")
            if not note_segments:
                continue

            gained = self._append_group_segments(group, note_segments, elements)
            current_time = start_time + duration_seconds

        metadata: dict[str, object] = {
            "feel": self.feel,
            "quantize_grid": self.quantize_grid,
        }
        if self.feel == "swing":
            metadata["swing_ratio"] = self.swing_ratio
        elif self.feel == "triplet":
            metadata["tuplet_division"] = 3
        elif self.feel == "quintuplet":
            metadata["tuplet_division"] = 5

        collapsed_elements = self._collapse_tuplets(elements, metadata)
        return Seq(elements=collapsed_elements, metadata=metadata)

    def _group_notes(self, notes: list[RecordedNote]) -> list[list[RecordedNote]]:
        groups: list[list[RecordedNote]] = []
        current_group: list[RecordedNote] = []
        current_start: float | None = None

        for note in notes:
            if (
                current_group
                and current_start is not None
                and abs(note.start_time - current_start)
                > CHORD_GROUPING_TOLERANCE_SECONDS
            ):
                groups.append(current_group)
                current_group = []
                current_start = None

            if not current_group:
                current_start = note.start_time
            current_group.append(note)

        if current_group:
            groups.append(current_group)

        return groups

    def _seconds_to_beats(self, seconds: float) -> float:
        return seconds * self.default_tempo / 60.0

    def _beats_to_seconds(self, beats: float) -> float:
        return beats * 60.0 / self.default_tempo

    def _grid_value(self) -> float:
        if self.feel == "triplet":
            return 1.0 / 3.0
        if self.feel == "quintuplet":
            return 0.2
        return self.quantize_grid

    def _quantize_beats(self, beats: float, *, kind: str) -> float:
        beats = max(beats, 0.0)
        grid = self._grid_value()

        if kind == "note" and self.feel == "swing":
            if SWING_DETECTION_MIN_BEATS < beats < SWING_DETECTION_MAX_BEATS:
                long = max(0.0, min(1.0, self.swing_ratio))
                short = max(0.0, 1.0 - long)
                target = long if self._swing_next_is_long else short
                self._swing_next_is_long = not self._swing_next_is_long
                return target
            self._swing_next_is_long = True

        if grid > 0:
            return round(beats / grid) * grid
        return beats

    def _segments_for_beats(
        self, beats: float, *, kind: str
    ) -> list[tuple[int, int, float]]:
        quantized = self._quantize_beats(beats, kind=kind)
        return self._segment_beats(max(quantized, 0.0))

    def _segment_beats(self, beats: float) -> list[tuple[int, int, float]]:
        segments: list[tuple[int, int, float]] = []
        remaining = beats
        grid = self._grid_value()
        tolerance = max(grid / 16.0 if grid else 0.005, 0.005)

        while remaining > tolerance:
            denom, dots = beats_to_duration(remaining)
            length = duration_value_to_beats(denom, dots)
            if length <= 0:
                break
            segments.append((denom, dots, length))
            remaining = max(0.0, remaining - length)

        return segments

    def _append_rest_segments(
        self, segments: list[tuple[int, int, float]], elements: list
    ) -> float:
        total = 0.0
        for denom, dots, length in segments:
            elements.append(Rest(duration=denom, dots=dots))
            total += length
        return total

    def _append_group_segments(
        self,
        group: list[RecordedNote],
        segments: list[tuple[int, int, float]],
        elements: list,
    ) -> float:
        total = 0.0
        is_chord = len(group) > 1
        note_infos = [
            midi_pitch_to_note(note.pitch) for note in group
        ]  # (letter, octave, accidentals)

        for idx, (denom, dots, length) in enumerate(segments):
            slur = idx < len(segments) - 1
            if is_chord:
                chord_notes = []
                for letter, octave, accidentals in note_infos:
                    accidental = accidentals[0] if accidentals else None
                    chord_notes.append(
                        Note(
                            pitch=letter,
                            duration=None,
                            dots=0,
                            octave=octave,
                            accidental=accidental,
                            slurred=slur,
                        )
                    )
                elements.append(
                    Chord(
                        notes=tuple(chord_notes),
                        duration=denom,
                        dots=dots,
                    )
                )
            else:
                letter, octave, accidentals = note_infos[0]
                accidental = accidentals[0] if accidentals else None
                elements.append(
                    Note(
                        pitch=letter,
                        duration=denom,
                        dots=dots,
                        octave=octave,
                        accidental=accidental,
                        slurred=slur,
                    )
                )
            total += length

        return total

    def _collapse_tuplets(self, elements: list, metadata: dict[str, object]) -> list:
        raw_division = metadata.get("tuplet_division")
        if not isinstance(raw_division, int):
            return elements

        if raw_division <= 1:
            return elements

        division = raw_division

        target_beats = 1.0 / division
        tolerance = target_beats / 8.0

        new_elements: list = []
        buffer: list = []
        buffer_beats = 0.0

        def flush_buffer() -> None:
            nonlocal buffer, buffer_beats
            if buffer:
                new_elements.extend(buffer)
            buffer = []
            buffer_beats = 0.0

        for elem in elements:
            beats = self._element_beats(elem)
            if beats is None or abs(beats - target_beats) > tolerance:
                flush_buffer()
                new_elements.append(elem)
                continue

            buffer.append(self._strip_slur(elem))
            buffer_beats += beats

            if len(buffer) == division:
                base_duration, base_dots = beats_to_duration(division * target_beats)
                if abs(buffer_beats - division * target_beats) <= tolerance * division:
                    new_elements.append(
                        Cram(
                            elements=list(buffer),
                            duration=base_duration,
                            dots=base_dots,
                        )
                    )
                else:
                    new_elements.extend(buffer)
                buffer = []
                buffer_beats = 0.0

        flush_buffer()
        return new_elements

    def _element_beats(self, element) -> float | None:
        if isinstance(element, Note) and element.duration is not None:
            return duration_value_to_beats(element.duration, element.dots)
        if isinstance(element, Rest) and element.duration is not None:
            return duration_value_to_beats(element.duration, element.dots)
        if isinstance(element, Chord) and element.duration is not None:
            return duration_value_to_beats(
                element.duration, getattr(element, "dots", 0)
            )
        return None

    @staticmethod
    def _strip_slur(element):
        if isinstance(element, Note) and element.slurred:
            return replace(element, slurred=False)
        if isinstance(element, Chord):
            stripped_notes = [
                replace(note, slurred=False) if note.slurred else note
                for note in element.notes
            ]
            return replace(element, notes=tuple(stripped_notes))
        return element

    def on_note(self, callback: Callable[[int, int, bool], None]) -> None:
        """Set a callback for note events.

        The callback receives (pitch, velocity, is_note_on).
        """
        self._on_note = callback


def transcribe(
    duration: float = 10.0,
    port_name: str | None = None,
    instrument: str = "piano",
    quantize_grid: float = 0.25,
    tempo: float = 120.0,
    feel: Literal["straight", "swing", "triplet", "quintuplet"] = "straight",
    swing_ratio: float = 2.0 / 3.0,
    on_note: Callable[[int, int, bool], None] | None = None,
    poll_interval: float = 0.01,
) -> "Score":  # noqa: F821
    """Record MIDI input and return a Score.

    This is a blocking function that records for the specified duration.

    Args:
        duration: Recording duration in seconds.
        port_name: MIDI input port name. If None, uses the first available port.
        instrument: Instrument name for the part.
        quantize_grid: Grid size in beats for quantization (0.25 = 16th notes).
        tempo: Tempo in BPM for duration calculations.
        feel: Quantization feel ("straight", "swing", "triplet", "quintuplet").
        swing_ratio: Portion of the beat allocated to the long swing note.
        on_note: Optional callback for note events (pitch, velocity, is_note_on).
        poll_interval: How often to poll for MIDI messages (seconds).

    Returns:
        A Score containing the recorded notes.

    Example:
        >>> from aldakit.midi.transcriber import transcribe
        >>> print("Recording for 10 seconds...")
        >>> score = transcribe(duration=10)
        >>> score.play()
    """

    session = TranscribeSession(
        port_name=port_name,
        quantize_grid=quantize_grid,
        default_tempo=tempo,
        feel=feel,
        swing_ratio=swing_ratio,
    )

    if on_note:
        session.on_note(on_note)

    session.start(port_name)

    # Record for the specified duration
    start = time.time()
    while time.time() - start < duration:
        session.poll()
        time.sleep(poll_interval)

    seq = session.stop()

    # Build a Score from the recorded notes

    elements = [Part(instruments=(instrument,)), Tempo(bpm=tempo)]
    elements.extend(seq.elements)

    return Score.from_elements(*elements)


def list_input_ports() -> list[str]:
    """List available MIDI input ports.

    Returns:
        List of port names.
    """
    observer = Observer()
    return [p.display_name for p in observer.get_input_ports()]



"""Convert MIDI sequences to AST nodes for import."""

from __future__ import annotations

from dataclasses import dataclass

from ..ast_nodes import (
    ChordNode,
    DurationNode,
    EventSequenceNode,
    LispListNode,
    LispNumberNode,
    LispSymbolNode,
    NoteLengthNode,
    NoteNode,
    OctaveSetNode,
    PartDeclarationNode,
    RestNode,
    RootNode,
)
from .types import INSTRUMENT_PROGRAMS, MidiNote, MidiSequence

# MIDI pitch to note letter and accidental
# We use sharps for black keys
PITCH_CLASS_TO_NOTE: list[tuple[str, list[str]]] = [
    ("c", []),  # 0
    ("c", ["+"]),  # 1 (C#)
    ("d", []),  # 2
    ("d", ["+"]),  # 3 (D#)
    ("e", []),  # 4
    ("f", []),  # 5
    ("f", ["+"]),  # 6 (F#)
    ("g", []),  # 7
    ("g", ["+"]),  # 8 (G#)
    ("a", []),  # 9
    ("a", ["+"]),  # 10 (A#)
    ("b", []),  # 11
]


# Standard note duration values and their lengths in quarter notes
# (duration_value, length_in_quarters)
DURATION_VALUES: list[tuple[int, float]] = [
    (1, 4.0),  # whole note
    (2, 2.0),  # half note
    (4, 1.0),  # quarter note
    (6, 0.6666666667),  # two-thirds beat (used for swing)
    (8, 0.5),  # eighth note
    (12, 0.3333333333),  # triplet eighth (12th note)
    (16, 0.25),  # sixteenth note
    (20, 0.2),  # quintuplet eighth (20th note)
    (24, 0.1666666667),  # triplet sixteenth (24th note)
    (32, 0.125),  # thirty-second note
    (40, 0.1),  # quintuplet sixteenth
    (48, 0.0833333333),  # triplet thirty-second
    (64, 0.0625),  # sixty-fourth note
    (80, 0.05),  # quintuplet thirty-second
]

# Dotted versions add 50% more
DOTTED_DURATION_VALUES: list[tuple[int, int, float]] = [
    (1, 1, 6.0),  # dotted whole
    (2, 1, 3.0),  # dotted half
    (4, 1, 1.5),  # dotted quarter
    (8, 1, 0.75),  # dotted eighth
    (12, 1, 0.5),  # dotted triplet eighth
    (16, 1, 0.375),  # dotted sixteenth
    (24, 1, 0.25),  # dotted triplet sixteenth
]


# Reverse mapping from GM program to instrument name
PROGRAM_TO_INSTRUMENT: dict[int, str] = {}
for name, program in INSTRUMENT_PROGRAMS.items():
    if program not in PROGRAM_TO_INSTRUMENT:
        PROGRAM_TO_INSTRUMENT[program] = name


@dataclass
class _QuantizedNote:
    """A note with quantized timing and duration."""

    pitch: int
    velocity: int
    start_beat: float
    duration_beats: float
    channel: int
    start_seconds: float
    duration_seconds: float


def midi_pitch_to_note(pitch: int) -> tuple[str, int, list[str]]:
    """Convert a MIDI pitch number to note name, octave, and accidentals.

    Args:
        pitch: MIDI pitch number (0-127, where 60 = middle C = C4).

    Returns:
        Tuple of (letter, octave, accidentals).
    """
    # MIDI note 60 = C4
    # octave = pitch // 12 - 1
    # C0 = MIDI 12
    octave = (pitch // 12) - 1
    pitch_class = pitch % 12

    letter, accidentals = PITCH_CLASS_TO_NOTE[pitch_class]
    return letter, octave, accidentals


def seconds_to_beats(seconds: float, bpm: float) -> float:
    """Convert seconds to beats at a given tempo."""
    return seconds * bpm / 60.0


def _make_duration_node(denominator: int, dots: int = 0) -> DurationNode:
    """Create a DurationNode with the given note length and dots."""
    return DurationNode(
        components=[NoteLengthNode(denominator=denominator, dots=dots, position=None)],
        position=None,
    )


def _make_tempo_node(bpm: float, global_: bool = False) -> LispListNode:
    """Create a tempo or tempo! lisp node."""
    symbol = "tempo!" if global_ else "tempo"
    return LispListNode(
        elements=[
            LispSymbolNode(name=symbol, position=None),
            LispNumberNode(value=int(round(bpm)), position=None),
        ],
        position=None,
    )


def beats_to_duration(beats: float) -> tuple[int, int]:
    """Convert beats to the closest duration value and dots.

    Args:
        beats: Duration in beats (quarter notes).

    Returns:
        Tuple of (duration_value, dots). Duration value is Alda's notation
        where 4 = quarter, 8 = eighth, etc.
    """
    candidate = beats
    expected = getattr(candidate, "expected", candidate)
    beats = float(expected)

    if beats <= 0:
        return 4, 0  # Default to quarter note

    # Try exact matches first
    for duration_value, length in DURATION_VALUES:
        if abs(beats - length) < 0.01:
            return duration_value, 0

    # Try dotted values
    for duration_value, dots, length in DOTTED_DURATION_VALUES:
        if abs(beats - length) < 0.01:
            return duration_value, dots

    # Find closest match
    best_duration = 4
    best_dots = 0
    best_diff = float("inf")

    for duration_value, length in DURATION_VALUES:
        diff = abs(beats - length)
        if diff < best_diff:
            best_diff = diff
            best_duration = duration_value
            best_dots = 0

    for duration_value, dots, length in DOTTED_DURATION_VALUES:
        diff = abs(beats - length)
        if diff < best_diff:
            best_diff = diff
            best_duration = duration_value
            best_dots = dots

    return best_duration, best_dots


def duration_value_to_beats(denominator: int, dots: int = 0) -> float:
    """Convert a duration value back to beats."""
    if denominator <= 0:
        return 0.0
    base = 4.0 / denominator
    total = base
    addition = base / 2.0
    for _ in range(dots):
        total += addition
        addition /= 2.0
    return total


def quantize_to_grid(value: float, grid: float) -> float:
    """Quantize a value to the nearest grid point."""
    if grid <= 0:
        return value
    return round(value / grid) * grid


def midi_to_ast(
    sequence: MidiSequence,
    *,
    quantize_grid: float = 0.25,  # Quantize to sixteenth notes by default
    default_bpm: float = 120.0,
) -> RootNode:
    """Convert a MidiSequence to an AST RootNode.

    Args:
        sequence: The MIDI sequence to convert.
        quantize_grid: Grid size in beats for quantization (0.25 = 16th notes).
        default_bpm: Default tempo if none specified in the sequence.

    Returns:
        A RootNode representing the imported MIDI.
    """
    # Determine tempo
    tempo_changes = sorted(sequence.tempo_changes, key=lambda t: t.time)
    bpm = tempo_changes[0].bpm if tempo_changes else default_bpm
    per_part_tempos = tempo_changes[1:] if tempo_changes else []
    tempo_events = [(tc.time, tc.bpm) for tc in per_part_tempos]

    # Group notes by channel
    channels: dict[int, list[MidiNote]] = {}
    for note in sequence.notes:
        if note.channel not in channels:
            channels[note.channel] = []
        channels[note.channel].append(note)

    # Get program changes to determine instruments
    channel_programs: dict[int, int] = {}
    for pc in sequence.program_changes:
        if pc.channel not in channel_programs:
            channel_programs[pc.channel] = pc.program

    # Build AST for each channel as a part
    children: list = []

    # Add tempo if not default
    if abs(bpm - 120.0) > 0.1:
        children.append(_make_tempo_node(bpm, global_=True))

    for channel in sorted(channels.keys()):
        notes = channels[channel]
        if not notes:
            continue

        # Determine instrument
        program = channel_programs.get(channel, 0)
        instrument_name = PROGRAM_TO_INSTRUMENT.get(program, "piano")

        # Create part declaration
        part_node = PartDeclarationNode(
            names=[instrument_name],
            alias=None,
            position=None,
        )
        children.append(part_node)

        # Convert notes to quantized form
        quantized = _quantize_notes(notes, bpm, quantize_grid)

        # Convert to AST events
        events = _notes_to_events(quantized, bpm, tempo_events)
        if events:
            event_seq = EventSequenceNode(events=events, position=None)
            children.append(event_seq)

    return RootNode(children=children, position=None)


def _quantize_notes(
    notes: list[MidiNote], bpm: float, grid: float
) -> list[_QuantizedNote]:
    """Quantize notes to a grid."""
    result = []

    for note in notes:
        start_beats = seconds_to_beats(note.start_time, bpm)
        duration_beats = seconds_to_beats(note.duration, bpm)

        # Quantize start time
        start_beats = quantize_to_grid(start_beats, grid)

        # Quantize duration to nearest standard value
        duration_beats = max(grid, quantize_to_grid(duration_beats, grid))

        result.append(
            _QuantizedNote(
                pitch=note.pitch,
                velocity=note.velocity,
                start_beat=start_beats,
                duration_beats=duration_beats,
                channel=note.channel,
                start_seconds=note.start_time,
                duration_seconds=note.duration,
            )
        )

    # Sort by start time, then by pitch
    result.sort(key=lambda n: (n.start_beat, n.pitch))
    return result


def _notes_to_events(
    notes: list[_QuantizedNote], bpm: float, tempo_events: list[tuple[float, float]]
) -> list:
    """Convert quantized notes to AST events."""
    if not notes:
        return []

    events: list = []
    current_beat = 0.0
    current_octave = 4  # Default octave
    tempo_index = 0

    i = 0
    while i < len(notes):
        note = notes[i]
        tempo_index = _emit_due_tempos(
            tempo_events, tempo_index, note.start_seconds, events
        )

        # Insert rest if there's a gap
        gap = note.start_beat - current_beat
        if gap > 0.01:
            rest_duration, rest_dots = beats_to_duration(gap)
            rest_node = RestNode(
                duration=_make_duration_node(rest_duration, rest_dots),
                position=None,
            )
            events.append(rest_node)
            current_beat = note.start_beat

        # Check for simultaneous notes (chord)
        chord_notes = [note]
        j = i + 1
        while j < len(notes) and abs(notes[j].start_beat - note.start_beat) < 0.01:
            chord_notes.append(notes[j])
            j += 1

        if len(chord_notes) > 1:
            # Create chord
            chord_elements = []
            chord_duration = chord_notes[0].duration_beats
            duration_val, dots = beats_to_duration(chord_duration)

            # Set octave for first note of chord if needed
            first_letter, first_octave, first_acc = midi_pitch_to_note(
                chord_notes[0].pitch
            )
            if first_octave != current_octave:
                events.append(OctaveSetNode(octave=first_octave, position=None))
                current_octave = first_octave

            for idx, cn in enumerate(chord_notes):
                letter, octave, accidentals = midi_pitch_to_note(cn.pitch)
                # Duration goes on first note only (Alda convention)
                note_duration = (
                    _make_duration_node(duration_val, dots) if idx == 0 else None
                )
                chord_elements.append(
                    NoteNode(
                        letter=letter,
                        accidentals=accidentals,
                        duration=note_duration,
                        slurred=False,
                        position=None,
                    )
                )

            chord_node = ChordNode(
                notes=chord_elements,
                position=None,
            )
            events.append(chord_node)
            current_beat = note.start_beat + chord_duration
            i = j

        else:
            # Single note
            letter, octave, accidentals = midi_pitch_to_note(note.pitch)
            duration_val, dots = beats_to_duration(note.duration_beats)

            # Set octave if changed
            if octave != current_octave:
                events.append(OctaveSetNode(octave=octave, position=None))
                current_octave = octave

            note_node = NoteNode(
                letter=letter,
                accidentals=accidentals,
                duration=_make_duration_node(duration_val, dots),
                slurred=False,
                position=None,
            )
            events.append(note_node)
            current_beat = note.start_beat + note.duration_beats
            i += 1

    _emit_due_tempos(tempo_events, tempo_index, float("inf"), events)
    return events


def _emit_due_tempos(
    tempo_events: list[tuple[float, float]],
    index: int,
    target_time: float,
    events: list,
) -> int:
    while index < len(tempo_events) and tempo_events[index][0] <= target_time + 1e-4:
        events.append(_make_tempo_node(tempo_events[index][1], global_=False))
        index += 1
    return index



from __future__ import annotations

from .choice_input import choice
from .dialogs import (button_dialog, checkboxlist_dialog, input_dialog,
                      message_dialog, progress_dialog, radiolist_dialog,
                      yes_no_dialog)
from .progress_bar import ProgressBar, ProgressBarCounter
from .prompt import (CompleteStyle, PromptSession, confirm,
                     create_confirm_session, prompt)
from .utils import (clear, clear_title, print_container, print_formatted_text,
                    set_title)

__all__ = [
    # Dialogs.
    "input_dialog",
    "message_dialog",
    "progress_dialog",
    "checkboxlist_dialog",
    "radiolist_dialog",
    "yes_no_dialog",
    "button_dialog",
    # Prompts.
    "PromptSession",
    "prompt",
    "confirm",
    "create_confirm_session",
    "CompleteStyle",
    # Progress bars.
    "ProgressBar",
    "ProgressBarCounter",
    # Choice selection.
    "choice",
    # Utils.
    "clear",
    "clear_title",
    "print_container",
    "print_formatted_text",
    "set_title",
]



"""MIDI generation and playback for Alda."""

from .backends import LibremidiBackend, MidiBackend
from .generator import MidiGenerator, generate_midi
from .types import (
    INSTRUMENT_PROGRAMS,
    GeneralMidiProgram,
    MidiControlChange,
    MidiNote,
    MidiProgramChange,
    MidiSequence,
    MidiTempoChange,
    note_to_midi,
)

__all__ = [
    # Types
    "MidiSequence",
    "MidiNote",
    "MidiProgramChange",
    "MidiControlChange",
    "MidiTempoChange",
    "GeneralMidiProgram",
    "INSTRUMENT_PROGRAMS",
    "note_to_midi",
    # Generator
    "MidiGenerator",
    "generate_midi",
    # Backends
    "MidiBackend",
    "LibremidiBackend",
]
